// Copyright (c) 2006 David Drysdale.  See license.txt for license info
import java.awt.*;

// Extension to the AWT Canvas class
class AnimCanvas extends Canvas implements Drawable2D {
  // Graphics vars
  double viewMinX_;
  double viewMaxX_;
  double viewMinY_;
  double viewMaxY_;
  Dimension size_; // Size in pixels of the Canvas
  Rectangle bounds_; // Useable area 
  double scale_ = 0.0;  // Scaling factor to convert coords to pixels
  Graphics g_;   // Clipped gc
  Image buff_[];
  int currentShow_ = 0;
  int currentDraw_ = 1;
  int lastx_;
  int lasty_;
  Control ctrlRoot_;
  ClickConsumer clickConsumer_;
  DragConsumer dragConsumer_;
  static final int xmargin_ = 5;
  static final int ymargin_ = 3;

  // Set the logical extents of the plotting canvas
  public AnimCanvas(double minx, double miny, double maxx, double maxy, DragConsumer drag) {
    viewMinX_ = minx;
    viewMinY_ = miny;
    viewMaxX_ = maxx;
    viewMaxY_ = maxy;
    dragConsumer_ = drag;
  }
  public void setDragConsumer(DragConsumer drag) {dragConsumer_ = drag;}
  public DragConsumer getDragConsumer() {return dragConsumer_;}
  public void setClickConsumer(ClickConsumer click) {clickConsumer_ = click;}
  public ClickConsumer getClickConsumer() {return clickConsumer_;}
  public void setControl(Control c) {ctrlRoot_ = c;}
  public Control getControl() {return ctrlRoot_;}
  public void paint(Graphics g) {
    if (scale_ == 0.0) {
      // First time in; now we know the size, do some sums
      size_ = getSize();
      // Size of graphics is [xmargin_, size_.width-xmargin_) x [ymargin_, size_.width-ymargin_)
      // Logical size is [ViewMinX_, ViewMaxX_) x [ViewMinY_, ViewMaxY_)
      double scalex = (size_.width - 2*xmargin_) / (viewMaxX_ - viewMinX_);
      double scaley = (size_.height - 2*ymargin_) / (viewMaxY_ - viewMinY_);
      scale_ = (scalex < scaley) ? scalex : scaley;
      // Figure out the boundary of the viewing area in pixel coords
      bounds_ = new Rectangle(xmargin_, 
                              ymargin_,
                              (int)((viewMaxX_ - viewMinX_) * scale_),
                              (int)((viewMaxY_ - viewMinY_) * scale_));

      // Create a pair of images for double buffering
      buff_ = new Image[2];
      buff_[0] = createImage(bounds_.width, bounds_.height);
      buff_[1] = createImage(bounds_.width, bounds_.height);
      // Flip the buffers twice to get a filled one to the front
      drawDone();
      drawDone();
    }
    // Fill the backdrop
    g.setColor(Color.black);
    g.fillRect(0,0,size_.width,size_.height);
    // Draw the extent of the usable area
    g.setColor(Color.red);
    g.drawRect(bounds_.x-1, bounds_.y-1, bounds_.width+1, bounds_.height+1);
    // Draw the current image
    update(g);
  }
  public void update(Graphics g) {
    // Draw the current image into the canvas
    g.drawImage(buff_[currentShow_], bounds_.x, bounds_.y, null);
  }

  public void drawDone() {
    // If there are controls, draw them on top
    if (ctrlRoot_ != null) {
      ctrlRoot_.display(this, g_);
    }

    // Flip the images
    currentShow_ = (1 - currentShow_);
    currentDraw_ = (1 - currentDraw_);

    // Clear the new drawing image
    g_ = buff_[currentDraw_].getGraphics();
    g_.setColor(Color.black);
    g_.fillRect(0,0,bounds_.width,bounds_.height);

    // Update the shown image
    update(this.getGraphics());
  }

  // Graphics scaling functions 
  private int x(double varx) {
    return (int)((varx-viewMinX_) * scale_);
  }
  private int y(double vary) {
    return (int)((vary-viewMinY_) * scale_);
  }
  // Graphics operations
  public void point(double xc, double yc, Color color) {
    g_.setColor(color);
    int xi = x(xc);
    int yi = y(yc);
    g_.fillOval(xi-1,yi-1,4,4);
  }
  public void point(Point2D p, Color color) {
    point(p.x, p.y, color);
  }
  protected void line(double x1, double y1, double x2, double y2, Color color) {
    g_.setColor(color);
    g_.drawLine(x(x1),y(y1),x(x2),y(y2));
  }
  protected void line(Line2D line) {
    if (Double.isInfinite(line.a.x) || Double.isInfinite(line.a.y) || 
        Double.isInfinite(line.b.x) || Double.isInfinite(line.b.y)) {
      return;
    } 
    if (Double.isNaN(line.a.x) || Double.isNaN(line.a.y) || 
        Double.isNaN(line.b.x) || Double.isNaN(line.b.y)) {
      return;
    } 
    line(line.a.x, line.a.y, line.b.x, line.b.y, line.c);
  }
  protected void lines(Line2D lines[]) {
    for (int ii=0; ii < lines.length; ii++) {
      line(lines[ii]);
    }
  }
  protected void circle(double xc, double yc, double radius, Color color) {
    // Bounding box for circle in logical coordinates is (xc-r, yc-r) -> (xc+r, yc+r)
    double x1 = xc-radius;
    double x2 = xc+radius;
    double y1 = yc - radius;
    double y2 = yc + radius;
    int x1i = x(x1);
    int x2i = x(x2);
    int y1i = y(y1);
    int y2i = y(y2);
    g_.setColor(color);
    int left = Math.min(x1i,x2i);
    int top = Math.min(y1i,y2i);
    int width = Math.abs(x1i-x2i);
    int height = Math.abs(y1i-y2i);
    g_.drawOval(left, top, width, height);
  }
  protected void fill(double x1, double y1, double x2, double y2, Color color) {
    int x1i = x(x1);
    int x2i = x(x2);
    int y1i = y(y1);
    int y2i = y(y2);
    g_.setColor(color);
    int left = Math.min(x1i,x2i);
    int top = Math.min(y1i,y2i);
    int width = Math.abs(x1i-x2i);
    int height = Math.abs(y1i-y2i);
    g_.fillRect(left, top, width, height);
  }
  protected void text(double xc, double yc, String str, Color color) {
    g_.setColor(color);
    Font f = new Font("TimesRoman", Font.PLAIN, 12);
    g_.setFont(f);
    FontMetrics fm = g_.getFontMetrics();
    g_.drawString(str, x(xc), y(yc)+fm.getAscent());
  }
  public boolean mouseDown(Event e, int x, int y) {
    lastx_ = x;
    lasty_ = y;
    if (clickConsumer_ != null) {
      clickConsumer_.click(viewMinX_ + ((double)x/scale_), viewMinY_ + ((double)y/scale_),x,y);
    }
    return true;
  }
  public boolean mouseUp(Event e, int x, int y) {
    if (clickConsumer_ != null) {
      clickConsumer_.release(viewMinX_ + ((double)x/scale_), viewMinY_ + ((double)y/scale_),x,y);
    }
    return true;
  }
  public boolean mouseDrag(Event e, int x, int y) {
    double deltax = (double)(x - lastx_)/(double)size_.width;
    double deltay = (double)(y - lasty_)/(double)size_.height;
    if (dragConsumer_ != null) {
      if ((e.modifiers & (Event.CTRL_MASK|Event.ALT_MASK|Event.META_MASK)) != 0) {
        dragConsumer_.ctrlDrag(deltax, deltay, x, y);
      } else {
        dragConsumer_.drag(deltax, deltay, x, y);
      }
    }
    lastx_ = x;
    lasty_ = y;
    return true;
  }
}

