// Copyright (c) 2006 David Drysdale.  See license.txt for license info
class Camera implements DragConsumer {
  // Thing to trigger on changes
  Triggerable trigger_;
  // Current position of the camera in world coordinates
  Point3D pos_;
  // Current direction of the camera.  This is not a unit vector; its length
  // gives the position of "centre of attention" (pos_+dir_)
  Point3D dir_; 
  // The "centre of attention", for convenience
  Point3D centre_;
  // Amount to scale the viewport by
  double scale_ = 1.0;
  // Whether to dump out copious trace info
  private boolean traceT_ = false;
  private boolean traceP_ = false;
  private boolean traceZ_ = false; 

  public Camera(Triggerable trigger, Point3D pos, Point3D dir, double scale) {
    trigger_ = trigger;
    pos_ = pos;
    dir_ = dir;
    centre_ = new Point3D();
    centre_.x = dir_.x + pos_.x;
    centre_.y = dir_.y + pos_.y;
    centre_.z = dir_.z + pos_.z;
    scale_ = scale;
  }
  public Camera(Triggerable trigger, Point3D pos, double scale) {
    trigger_ = trigger;
    pos_ = pos;
    dir_ = new Point3D();
    centre_ = new Point3D();
    centre_.x = dir_.x + pos_.x;
    centre_.y = dir_.y + pos_.y;
    centre_.z = dir_.z + pos_.z;
    scale_ = scale;
  }
  public boolean drag(double x, double y, int rawx, int rawy) {
    pan(x, y);
    return true; // Always consume the event
  }
  public boolean ctrlDrag(double x, double y, int rawx, int rawy) {
    zoom(x, y);
    return true; // Always consume the event
  }
  public void setDir(Point3D centre) {
    centre_ = centre;
    dir_.x = centre_.x - pos_.x;
    dir_.y = centre_.y - pos_.y;
    dir_.z = centre_.z - pos_.z;
    if (traceT_) System.out.println("Camera is at ("+pos_.x+","+pos_.y+","+pos_.z+") pointing in direction " + 
                                   "("+dir_.x+","+dir_.y+","+dir_.z+")");
  }
  public void setDir(Line3DSequence frames) {
    centre_ = frames.findCentre();
    setDir(centre_);
  }
  public void pan(double x, double y) {
    // Use x and y to move the camera on a sphere centered at centre=(pos_+dir_)
    // The distances from pos_ to camera stays fixed, and the direction of the
    // camera is rotated so that it continues to point at the centre
    Point3D pos = new Point3D(pos_);
    double d = dir_.length();
    // X-rotation: vector in XY plane perpendicular to dir_ is (-dir_.y, dir_.x,0)
    Point3D rotDirX = new Point3D(-dir_.y, dir_.x, 0);
    rotDirX.makeUnit();
    double SCALE = 300;
    pos.x += rotDirX.x * d * SCALE * x;
    pos.y += rotDirX.y * d * SCALE * x;
    // noop: pos.z += rotDir.z * d * SCALE * x;

    // Y-rotation: need a unit vector that's perpendicular to both dir_ and also the unit
    // vector used in the X-rotation.  We can generate such a vector using the cross product
    //   (-dir_.y, dir_.x, 0) x 
    //   (dir_.x,  dir_.y, dir_.z) 
    // = (xz, yz, -(xx+yy))
    Point3D rotDirY = new Point3D();
    rotDirY.x = dir_.x*dir_.z;
    rotDirY.y = dir_.y*dir_.z; 
    rotDirY.z = -(dir_.x*dir_.x + dir_.y*dir_.y);
    rotDirY.makeUnit();

    // We threshold the z-position of the camera so that it cannot go 'over the top' or 
    // 'under the bottom'. sin(angle with horiz) = deltaz / oldd
    double deltaz = Math.abs(pos.z + SCALE*y*rotDirY.z - centre_.z);
    double oldd = dir_.length();
    double SINTHRESH = 0.97;
    if ((deltaz/oldd) < SINTHRESH) {
      pos.x += SCALE*y*rotDirY.x;
      pos.y += SCALE*y*rotDirY.y;
      pos.z += SCALE*y*rotDirY.z;
    }

    // Make sure we've preserved distance -- move along the line from the new
    // camera pos_ to the centre_ to get the old distance back
    dir_.x = centre_.x - pos.x;
    dir_.y = centre_.y - pos.y;
    dir_.z = centre_.z - pos.z;
    double newd = dir_.length();
    double changed = newd - oldd;
    pos.x += dir_.x * (changed/newd);
    pos.y += dir_.y * (changed/newd);
    pos.z += dir_.z * (changed/newd);

    if (traceP_) System.out.println("pan ("+pos_.x+","+pos_.y+","+pos_.z+") --" +
                                   "["+x+","+y+"]--> " +
                                   "("+pos.x+","+pos.y+","+pos.z+")");
    pos_ = pos;
    setDir(centre_);
    trigger_.trigger();
  }
  public void zoom(double x, double y) {
    // Work out the current camera->centre distance
    double d = dir_.length();

    // X-zoom scales the projection plane
    double oldscale = scale_;
    scale_ += (x+y) * scale_;
    if (traceZ_) System.out.println("Input x+y="+(x+y)+" changes scale_ from "+oldscale+" to "+scale_);

    if (0==1) {
      // Disable this; now that the projection plane is a *proportion* of the distance to the centre,
      // moving the camera in and out has no effect
      // Y-zoom moves the camera
      double SCALE=50;
      double yscaled = y * SCALE;
      double MAXMOVE = 0.90;
      if (yscaled < MAXMOVE) {
        // Only do the move if it's not going to take the camera to the other side
        pos_.x += dir_.x * yscaled;
        pos_.y += dir_.y * yscaled;
        pos_.z += dir_.z * yscaled;
        setDir(centre_);
        if (traceZ_) System.out.println("Input y="+y+" changes camera-centre dist from "+d+" to "+dir_.length());
      }
    }
    trigger_.trigger();
  }
  public Line2DSequence transform(Line3DSequence frames) {
    Line2DSequence result = new Line2DSequence(frames.length(), frames.numLines_);
    for (int ii=0; ii< frames.length(); ii++) {
      result.deltat_[ii] = frames.deltat_[ii];
      result.lines_[ii] = transform(frames.lines_[ii]);
    }
    return result;
  }
  public Line2D[] transform(Line3D[] lines) {
    Line2D[] result = new Line2D[lines.length];
    for (int ii=0; ii<lines.length; ii++) {
      result[ii] = transform(lines[ii]);
    }
    return result;
  }
  public Line2D transform(Line3D line) {
    Line2D result = new Line2D();
    result.a = transform(line.a);
    result.b = transform(line.b);
    result.c = line.c;
    return result;
  }
  public Point2D[] transform(Point3D[] points) {
    Point2D[] result = new Point2D[points.length];
    for (int ii = 0; ii < points.length; ii++) {
      result[ii] = transform(points[ii]);
    }
    return result;
  }
  public Point2D transform(Point3D point) {
    // First step is to make the point relative to the camera (i.e. to make the camera
    // the origin of a new coordinate system)
    Point3D pt = new Point3D(point);
    pt.x -= pos_.x;
    pt.y -= pos_.y;
    pt.z -= pos_.z;
    if (traceT_) System.out.println("translate ("+point.x+","+point.y+","+point.z+")->" +
                                    "("+pt.x+","+pt.y+","+pt.z+")");

    // Now we rotate the coordinate system so that the direction of the camera becomes
    // the z-axis in a new coordinate system.  

    // First rotate in the XY plane so that the x-coordinate of dir_ gets zeroed
    //  sin A = (x/d);  cos A = (y/d);  d^2 = x^2 + y^2
    double d = Math.sqrt(dir_.x*dir_.x+dir_.y*dir_.y);
    Point3D pt2 = new Point3D();
    pt2.x = pt.x*(dir_.y/d) - pt.y*(dir_.x/d);
    pt2.y = pt.x*(dir_.x/d) + pt.y*(dir_.y/d);
    pt2.z = pt.z;
    if (traceT_) System.out.println("rotate ("+pt.x+","+pt.y+","+pt.z+")->" +
                                    "("+pt2.x+","+pt2.y+","+pt2.z+") with " +
                                    "("+(pos_.y/d)+" "+(-pos_.x/d)+" 0," +
                                    (pos_.x/d)+" "+(pos_.y/d)+" 0, 0 0 1)");

    // Second rotate in the YZ plane so that the y coordinate of dir_ gets zeroed
    //  sin B = (z/D);  cos B = (d/D); D^2 = d^2 + z^2
    double D = Math.sqrt(d*d+dir_.z*dir_.z);
    Point3D pt3 =  new Point3D();
    pt3.x = pt2.x;
    pt3.y = pt2.y*(dir_.z/D) - pt2.z*(d/D);
    pt3.z = pt2.y*(d/D) + pt2.z*(dir_.z/D);
    if (traceT_) System.out.println("rotate ("+pt2.x+","+pt2.y+","+pt2.z+")->" +
                                    "("+pt3.x+","+pt3.y+","+pt3.z+") with " +
                                   "(1 0 0, 0"+(pos_.y/d)+" "+(-d/D)+", 0 " +
                                   (d/D)+" "+(dir_.z/D)+")");
     
    // Sanity check: if the pt->pt2->pt3 rotations above are applied to dir_, should get:
    //  (dir_.x, dir_.y, dir_.z) -> (0, d, dir_.z) -> (0, 0, D)

    // If the point has a z coordinate less than dist/3 in this latest coordinate system, it is 
    // behind the camera or between the camera and the projection plane
    Point2D result = new Point2D();
    double dist = dir_.length();
    if (pt3.z < (dist/3)) {
      result.x = Double.POSITIVE_INFINITY;
      result.y = Double.POSITIVE_INFINITY;
    } else {
      // Project onto the plane
      result.x = scale_ * (dist/3) * (pt3.x/pt3.z);
      result.y = scale_ * (dist/3) * (pt3.y/pt3.z);
    }
    if (traceT_) System.out.println("project ("+pt3.x+","+pt3.y+","+pt3.z+")->" +
                                    "["+result.x+","+result.y+"]\n");
    return result;
  }
  public void dump() {
    if (!traceT_ && !traceP_ && !traceZ_) return;
    pos_.dump("[Camera] pos_ ");
    dir_.dump("[Camera] dir_ ");
    System.out.println("[Camera] dir_ length="+dir_.length());
    centre_.dump("[Camera] centre_ ");
  }
};