// Copyright (c) 2006 David Drysdale.  See license.txt for license info
class CompoundMove extends Move {
  String[] contents_;
  static boolean verbose_ = false;
  static final double EPSILON=0.001;
  static final char POSCHAR = '@';
  static final char DIRCHAR = '?';

  CompoundMove(String name, String[] names) {
    super(name);
    Move.register(name_, this);
    contents_ = new String[names.length];
    for (int ii=0; ii<names.length; ii++) {
      contents_[ii] = names[ii];
    }
  }

  private String[] getParts(String name) {
    // Break the string down at vertical bars
    int count = 1;
    int pos = 0;
    while (pos != -1) {
      pos = name.indexOf('|', pos);
      if (pos != -1) {
        // Found a bar
        count++;
        pos++; // Move past the bar for next time round
      }
    }
    if (verbose_) System.out.println("Name "+name+" has "+count+" parts");
    
    String[] parts = new String[count];
    count = 0;
    pos = 0;
    while (count < parts.length) {
      int nextpos = name.indexOf('|', pos);
      if (nextpos == -1) {
        parts[count] = name.substring(pos);
      } else {
        parts[count] = name.substring(pos, nextpos);
      }
      if (verbose_) System.out.println(" part["+count+"] = \""+parts[count]+"\"");
      count++;
      pos = nextpos+1;
    }
    return parts;
  }
  private Point3D getOffset(String partname) {
    Point3D offset = new Point3D();
    // Scan for marker followed by offset
    int pos = partname.indexOf(POSCHAR);
    if (pos != -1) {
      String offsetString = partname.substring(pos+1);
      offset.fromString(offsetString);
    }
    return offset;
  }
  private Point3D getOrient(String partname) {
    Point3D offset = new Point3D();
    // Scan for marker followed by offset
    int pos = partname.indexOf(DIRCHAR);
    if (pos != -1) {
      String offsetString = partname.substring(pos+1);
      offset.fromString(offsetString);
    }
    return offset;
  }
  private String getPartName(String partname) {
    // Scan for @ or :
    int pos1 = partname.indexOf(POSCHAR);
    int pos2 = partname.indexOf(DIRCHAR);
    if (pos1 != -1) {
      if (pos2 != -1) {
        return partname.substring(0, Math.min(pos1, pos2));
      } else {
        return partname.substring(0, pos1);
      }
    } else {
      if (pos2 != -1) {
        return partname.substring(0, pos2);
      } else {
        return partname;
      }
    }
  }
  
  public FiguresSequence getFigSeq() {
    FiguresSequence[] seqs = new FiguresSequence[contents_.length];
    // Figure out overall size
    int numframes=0;
    int numfigs=-1;
    for (int component=0; component<contents_.length; component++) {
      // Break this name into its parts
      String[] parts = getParts(contents_[component]);
      int thisNumFigs = 0;
      FiguresSequence[] partSeqs = new FiguresSequence[parts.length];
      int numFrames = -1;
      for (int part=0; part<parts.length; part++) {
        // Break the part into move and offset
        String partname = getPartName(parts[part]);
        Point3D offset = getOffset(parts[part]);
        Point3D orient = getOrient(parts[part]);
        Move partMove = Move.lookup(partname);
        // Get the figure sequence for the part and offset it
        if (verbose_) System.out.println("copy construct partSeqs[part="+part+"] from partmove=\""+partname+"\" at "+offset+" dir "+orient);
        partSeqs[part] = new FiguresSequence(partMove.getFigSeq(), offset, orient);
        if (verbose_) System.out.println("it has numFrames="+partSeqs[part].length()+", numFigs="+partSeqs[part].numFigs_);
        thisNumFigs += partSeqs[part].numFigs_;
        if (numFrames == -1) {
          numFrames = partSeqs[part].length();
        } else if (numFrames != partSeqs[part].length()) {
          throw new ArrayIndexOutOfBoundsException("mismatch in number of frames, "+numFrames+" vs. "+partSeqs[part].length());
        }
      }
      // Build a composite figure sequence for this name's parts
      seqs[component] = new FiguresSequence(numFrames, thisNumFigs);
      if (verbose_) System.out.println("seqs[component] has numFrames="+seqs[component].length()+", numFigs="+seqs[component].numFigs_);
      int resultFig = 0;
      for (int part=0; part<parts.length; part++) {
        for (int fig=0; fig<partSeqs[part].numFigs_; fig++, resultFig++) {
          seqs[component].color[resultFig] = partSeqs[part].color[fig];
          for (int frame = 0; frame < numFrames; frame++) {
            if (part == 0) {
              // Get the time offsets from the first part
              seqs[component].deltat_[frame] = partSeqs[0].deltat_[frame];
            } else {
              // The time offsets must also be the same
              if (partSeqs[part].deltat_[frame] != partSeqs[0].deltat_[frame]) {
                throw new ArrayIndexOutOfBoundsException("mismatch in frame timings at frame "+frame+
                                                         " of component "+component+", part "+part);
              }
            }
            if (verbose_) System.out.println("Setting seqs["+component+"].figure[frame="+frame+"][fig="+resultFig+"] "+ 
                                             "from partSeqs[part="+part+"].figure[frame="+frame+"][fig="+fig+"]");
            seqs[component].figure[frame][resultFig] = partSeqs[part].figure[frame][fig];
            seqs[component].figure[frame][resultFig] = partSeqs[part].figure[frame][fig];
          }
        }
      }

      numframes += seqs[component].length();
      if (numfigs == -1) {
        numfigs = seqs[component].numFigs_;
      } else  if (seqs[component].numFigs_ != numfigs) {
        throw new ArrayIndexOutOfBoundsException("mismatch in number of figures, "+numfigs+" vs. "+seqs[component].numFigs_);
      }
    }

    // Allocate the overall sequence
    FiguresSequence result = new FiguresSequence(numframes, numfigs);
    Point3D[] position = new Point3D[numfigs];
    Point3D[] direction = new Point3D[numfigs];
    for (int fig=0; fig<numfigs; fig++) {
      position[fig] = new Point3D();
      direction[fig] = new Point3D();
    }
    // Set up some dreibiens 
    Point3D[] i = new Point3D[numfigs];
    Point3D[] j = new Point3D[numfigs];
    Point3D[] k = new Point3D[numfigs];
    for (int fig=0; fig<numfigs; fig++) {
      i[fig] = new Point3D(1,0,0);
      j[fig] = new Point3D(0,1,0);
      k[fig] = new Point3D(0,0,1);
    }

    // Now fill it in
    int finalframe = 0;
    Point3D delta = new Point3D();
    // Take colors from the first component
    for (int fig=0; fig<numfigs; fig++) {
      result.color[fig] = seqs[0].color[fig];
    }
    // Loop over all component subsequences
    for (int component=0; component<contents_.length; component++) {
      // Loop over all of the frames in the current subsequence
      for (int frame=0; frame<seqs[component].length(); frame++,finalframe++) {
        if (verbose_) System.out.println("result frame["+finalframe+"]:");
        result.deltat_[finalframe] = seqs[component].deltat_[frame];
        // When joining subsequences, don't have any time between them
        if (frame==0) result.deltat_[finalframe] = EPSILON;
        // Loop over all the figures in this frame
        for (int fig=0; fig<numfigs; fig++) {
          if ((frame == 0) && (finalframe != 0)) {
            // This is the first frame of a subsequence (other than the first).  
            // We need to figure out the running delta that will make the position
            // and direction of this subsequence line up with the previous
            // subsequence.  This running delta will apply to all of the 
            // frames in the subsequence
            if (verbose_) System.out.println("start position of subsequence is "+
                                             seqs[component].figure[0][fig].posx_.doubleValue() + " " +
                                             seqs[component].figure[0][fig].posy_.doubleValue() + " " +
                                             seqs[component].figure[0][fig].posz_.doubleValue());
            position[fig].x -= seqs[component].figure[0][fig].posx_.doubleValue();
            position[fig].y -= seqs[component].figure[0][fig].posy_.doubleValue();
            position[fig].z -= seqs[component].figure[0][fig].posz_.doubleValue();
            if (verbose_) position[fig].dump("delta for fig " + fig + " is ");
            direction[fig].x -= seqs[component].figure[0][fig].lr_.doubleValue();
            direction[fig].y -= seqs[component].figure[0][fig].ud_.doubleValue();
            direction[fig].z -= seqs[component].figure[0][fig].tilt_.doubleValue();
          }

          // Copy across the FigurePosition
          result.figure[finalframe][fig] = new FigurePosition(seqs[component].figure[frame][fig]);
          if (frame == 0) {
            // Adjust position by applying the overall position change position[fig]
            result.figure[finalframe][fig].posx_.addDoubleValue(position[fig].x);
            result.figure[finalframe][fig].posy_.addDoubleValue(position[fig].y);
            result.figure[finalframe][fig].posz_.addDoubleValue(position[fig].z);
          } else {
            // Figure out where the figure is in this frame relative to the previous frame
            delta.x = seqs[component].figure[frame][fig].posx_.doubleValue() 
              - seqs[component].figure[frame-1][fig].posx_.doubleValue();
            delta.y = seqs[component].figure[frame][fig].posy_.doubleValue() 
              - seqs[component].figure[frame-1][fig].posy_.doubleValue();
            delta.z = seqs[component].figure[frame][fig].posz_.doubleValue() 
              - seqs[component].figure[frame-1][fig].posz_.doubleValue();
            
            // Adjust position by applying the overall position change position[fig] and 
            // also the difference between this frame and the previous frame
            result.figure[finalframe][fig].posx_.setDoubleValue(
              seqs[component].figure[frame-1][fig].posx_.doubleValue() + position[fig].x + 
              delta.x * i[fig].x + delta.y * j[fig].x + delta.z * k[fig].x);
            result.figure[finalframe][fig].posy_.setDoubleValue(
              seqs[component].figure[frame-1][fig].posy_.doubleValue() + position[fig].y + 
              delta.x * i[fig].y + delta.y * j[fig].y + delta.z * k[fig].y);
            result.figure[finalframe][fig].posz_.setDoubleValue(
              seqs[component].figure[frame-1][fig].posz_.doubleValue() + position[fig].z + 
              delta.x * i[fig].z + delta.y * j[fig].z + delta.z * k[fig].z);
          }
          
          // Adjust orientation 
          result.figure[finalframe][fig].lr_.addDoubleValue(direction[fig].x);
          result.figure[finalframe][fig].ud_.addDoubleValue(direction[fig].y);
          result.figure[finalframe][fig].tilt_.addDoubleValue(direction[fig].z);

          if (frame == (seqs[component].length()-1)) {
            // This is the last frame of the subsequence; remember the 
            // position and direction ready for the next subsequence
            position[fig].x = result.figure[finalframe][fig].posx_.doubleValue();
            position[fig].y = result.figure[finalframe][fig].posy_.doubleValue();
            position[fig].z = result.figure[finalframe][fig].posz_.doubleValue();
            if (verbose_) position[fig].dump("last frame of subseq ends with fig " + fig + " at ");
            direction[fig].x = result.figure[finalframe][fig].lr_.doubleValue();
            direction[fig].y = result.figure[finalframe][fig].ud_.doubleValue();
            direction[fig].z = result.figure[finalframe][fig].tilt_.doubleValue();
            // Also figure out the change in orientation
            double lr = (seqs[component].figure[frame][fig].lr_.doubleValue() -
                         seqs[component].figure[0][fig].lr_.doubleValue());
            double ud = (seqs[component].figure[frame][fig].ud_.doubleValue() -
                         seqs[component].figure[0][fig].ud_.doubleValue());
            double tilt = (seqs[component].figure[frame][fig].tilt_.doubleValue() -
                           seqs[component].figure[0][fig].tilt_.doubleValue());
            Matrix3D t = FigurePosition.getTransformMatrix(lr, ud, tilt);
            i[fig] = t.multiply(i[fig]);
            j[fig] = t.multiply(j[fig]);
            k[fig] = t.multiply(k[fig]);
            i[fig].makeUnit();
            j[fig].makeUnit();
            k[fig].makeUnit();
          }
        }
      }
    }
    return result;
  }
  
  public String toString() {
    String result = super.toString() + " type=compound contents["+contents_.length+"]=";
    for (int component=0;  component<contents_.length; component++) {
      if (component!=0) result+= " + ";
      result += "\"" + contents_[component] + "\"";
    }
    return result + "\n";
  }
  public char[] fromString(char[] s) {
    s = super.fromString(s);
    s = Format.expect(s, "type=compound");
    s = Format.expect(s, "contents[");
    // look for the ']'
    int jj = 0;
    while (s[jj] != ']') jj++;
    String ns = new String(s, 0, jj);
    int numconts = Integer.parseInt(ns, 10);
    contents_ = new String[numconts];
    s = Format.expect(s, jj, "]=");
    for (int component=0;  component<numconts; component++) {
      if (component!=0) s = Format.expect(s, "+");
      s = Format.expect(s,"\"");
      jj = 0;
      while (s[jj] != '\"') jj++;
      contents_[component] = new String(s, 0, jj);
      s = Format.expect(s, jj, "\"");
    }
    return s;
  }
  public String fromString(String s) {
    char[] chars = s.toCharArray();
    return new String(fromString(chars));
  }
};