// Copyright (c) 2006 David Drysdale.  See license.txt for license info
import java.awt.Color;

class FiguresSequence extends Sequence {
  public int numFigs_;
  public FigurePosition figure[][];
  public Color color[];
  
  FiguresSequence(int numFrames, int numFigs) {
    super(numFrames);
    numFigs_ = numFigs;
    figure = new FigurePosition[numFrames][numFigs];
    color = new Color[numFigs];
    Point3D zeroPoint = new Point3D(0,0,0);
    for (int ii=0; ii<numFrames; ii++) {
      for (int jj=0; jj<numFigs; jj++) {
        figure[ii][jj] = new FigurePosition(zeroPoint, 0, 0, 0);
      }
    }
  }
  FiguresSequence(FiguresSequence other, Point3D offset, Point3D orient) {
    super(other);
    int numFrames = other.length();
    numFigs_ = other.numFigs_;
    color = new Color[numFigs_];
    for (int jj=0; jj<numFigs_; jj++) {
      color[jj] = other.color[jj];
    }

    figure = new FigurePosition[numFrames][numFigs_];
    // Set up a dreibien (same for all figs)
    Point3D i = new Point3D(1,0,0);
    Point3D j = new Point3D(0,1,0);
    Point3D k = new Point3D(0,0,1);
    Matrix3D t = FigurePosition.getTransformMatrix(orient.x, orient.y, orient.z);
    i = t.multiply(i);
    j = t.multiply(j);
    k = t.multiply(k);
    i.makeUnit();
    j.makeUnit();
    k.makeUnit();

    Point3D delta = new Point3D();
    for (int frame=0; frame<numFrames; frame++) {
      for (int fig=0; fig<numFigs_; fig++) {
        figure[frame][fig] = new FigurePosition(other.figure[frame][fig]);
        if (frame == 0) {
          // Adjust position by applying the overall position change offset
          figure[frame][fig].posx_.addDoubleValue(offset.x);
          figure[frame][fig].posy_.addDoubleValue(offset.y);
          figure[frame][fig].posz_.addDoubleValue(offset.z);
        } else {
          // Figure out where the figure is in this frame relative to the previous frame
          delta.x = (other.figure[frame][fig].posx_.doubleValue() -
                     other.figure[frame-1][fig].posx_.doubleValue());
          delta.y = (other.figure[frame][fig].posy_.doubleValue() - 
                     other.figure[frame-1][fig].posy_.doubleValue());
          delta.z = (other.figure[frame][fig].posz_.doubleValue() - 
                     other.figure[frame-1][fig].posz_.doubleValue());
          
          // Adjust position by applying the overall position change offset and 
          // also the difference between this frame and the previous frame
          figure[frame][fig].posx_.setDoubleValue(
            other.figure[frame-1][fig].posx_.doubleValue() + offset.x + 
            delta.x * i.x + delta.y * j.x + delta.z * k.x);
          figure[frame][fig].posy_.setDoubleValue(
            other.figure[frame-1][fig].posy_.doubleValue() + offset.y + 
            delta.x * i.y + delta.y * j.y + delta.z * k.y);
          figure[frame][fig].posz_.setDoubleValue(
            other.figure[frame-1][fig].posz_.doubleValue() + offset.z + 
            delta.x * i.z + delta.y * j.z + delta.z * k.z);
        }
          
        // Adjust orientation 
        figure[frame][fig].lr_.addDoubleValue(orient.x);
        figure[frame][fig].ud_.addDoubleValue(orient.y);
        figure[frame][fig].tilt_.addDoubleValue(orient.z);
      }
    }
  }

  public void addFrame(int frame) {
    int numFrames = length();
    if (frame>=numFrames) return;
    super.addFrame(frame);
    FigurePosition[][] newfigure = new FigurePosition[numFrames+1][numFigs_];
    int ii = 0;
    int jj = 0;
    while (ii<numFrames) {
      for (int kk=0;kk<numFigs_;kk++) newfigure[jj][kk] = figure[ii][kk];
      jj++;
      if (ii == frame) {
        // insert an extra
        for (int kk=0;kk<numFigs_;kk++) newfigure[jj][kk] = figure[ii][kk];
        jj++;
      } 
      ii++;
    }
    figure = newfigure;
  }
  public void delFrame(int frame) {
    int numFrames = length();
    if (frame>=numFrames) return;
    super.delFrame(frame);
    FigurePosition[][] newfigure = new FigurePosition[numFrames-1][numFigs_];
    int ii = 0;
    int jj = 0;
    while (ii<numFrames) {
      if (ii != frame) {
        for (int kk=0; kk<numFigs_; kk++) newfigure[jj][kk] = figure[ii][kk];
        jj++;
      }
      ii++;
    }
    figure = newfigure;
  }
  Line3DSequence getLines3D(boolean includeBB, boolean includeAxes) {
    int count = numFigs_*Figure.lineCount_;
    if (includeBB) count += numFigs_*Figure.bbLineCount_;
    if (includeAxes) count += 3;

    Line3DSequence frames3d = new Line3DSequence(length(), 3+count);
    for (int frame = 0;  frame < frames3d.length(); frame++) {
      int line = 0;
      // Copy across the time delta
      frames3d.deltat_[frame] = deltat_[frame];
      // Add some axes, if so asked
      if (includeAxes) {
        frames3d.lines_[frame][line].a.x = 0; frames3d.lines_[frame][line].a.y = 0; frames3d.lines_[frame][line].a.z = 0; 
        frames3d.lines_[frame][line].b.x = 0; frames3d.lines_[frame][line].b.y = 0; frames3d.lines_[frame][line].b.z = 1; 
        frames3d.lines_[frame][line].c = Color.red; 
        line++;
        frames3d.lines_[frame][line].a.x = 0; frames3d.lines_[frame][line].a.y = 0; frames3d.lines_[frame][line].a.z = 0; 
        frames3d.lines_[frame][line].b.x = 0; frames3d.lines_[frame][line].b.y = 1; frames3d.lines_[frame][line].b.z = 0; 
        frames3d.lines_[frame][line].c = Color.white; 
        line++;
        frames3d.lines_[frame][line].a.x = 0; frames3d.lines_[frame][line].a.y = 0; frames3d.lines_[frame][line].a.z = 0; 
        frames3d.lines_[frame][line].b.x = 1; frames3d.lines_[frame][line].b.y = 0; frames3d.lines_[frame][line].b.z = 0; 
        frames3d.lines_[frame][line].c = Color.blue; 
        line++;
      }
      // Add in all the figures themselves
      for (int fig=0; fig<numFigs_; fig++) {
        Line3D[] figlines = figure[frame][fig].getLines();
        for (int kk = 0; kk < figlines.length; kk++) {
          frames3d.lines_[frame][line] = figlines[kk];
          frames3d.lines_[frame][line].c = color[fig];
          line++;
        }
      }
      // Add in the bounding boxes of the figures, if so asked
      if (includeBB) {
        for (int fig=0; fig<numFigs_; fig++) {
          Line3D[] figbox = figure[frame][fig].getBoundingBox();
          for (int kk = 0; kk < figbox.length; kk++) {
            frames3d.lines_[frame][line] = figbox[kk];
            line++;
          }
        }
      }
    }
    return frames3d;
  }
  public String toString() {
    String result = super.toString() + "\n";
    result += "figures[" + numFigs_ + "]=\n";
    for (int fig=0; fig<numFigs_; fig++) {
      int rgb = (color[fig].getRGB() & 0xFFFFFF);
      result += "figure "+fig+" color="+Integer.toHexString(rgb)+" {\n";
      for (int frame=0; frame<length(); frame++) {
        result += "frame "+frame+" {\n";
        result += figure[frame][fig];
        result += "}\n";
      }
      result += "}\n";
    }
    return result;
  }
  public char[] fromString(char[] s) {
    s = super.fromString(s); // fill out Sequence
    s = Format.expect(s, "figures[");
    // look for the ']'
    int jj = 0;
    while (s[jj] != ']') jj++;
    String ns = new String(s, 0, jj);
    numFigs_ = Integer.parseInt(ns, 10);
    // we now know both sizes, so allocate
    figure = new FigurePosition[length()][numFigs_];
    color = new Color[numFigs_];
    // move past ']='
    s = Format.expect(s, jj, "]=");
    for (int fig=0; fig<numFigs_; fig++) {
      s = Format.expect(s, "figure "+fig+" color=");

      int ii=0;
      while (Format.isWhitespace(s[ii])) ii++;
      jj=ii;
      while (Format.isHexDigit(s[jj])) jj++;
      String cs = new String(s, ii, jj-ii);
      int rgb = Integer.parseInt(cs, 16);
      color[fig] = new Color(rgb);

      s = Format.expect(s, jj, "{");
      for (int frame=0; frame<length(); frame++) {
        s = Format.expect(s, "frame "+frame+" {");
        figure[frame][fig] = new FigurePosition(new Point3D(),0,0,0);
        s = figure[frame][fig].fromString(s);
        s = Format.expect(s, "}");
      }
      s = Format.expect(s, "}");
    }
    return s;
  }
  public String fromString(String s) {
    char[] chars = s.toCharArray();
    return new String(fromString(chars));
  }
};