// Copyright (c) 2006 David Drysdale.  See license.txt for license info
import java.awt.Color;
import java.awt.Label;

class JointSelector implements ClickConsumer {
  ClickConsumer next_;
  Triggerable trigger_;
  Camera camera_;
  Label label_;
  boolean selected_ = false;
  FigurePosition[] figures_;
  Color[] colors_;
  AnimCanvas canvas_;
  static boolean verbose_ = false;

  JointSelector(Triggerable trigger, AnimCanvas canvas, Camera camera, Label label,
                FigurePosition[] figures, Color[] colors) {
    trigger_ = trigger;
    canvas_ = canvas;
    camera_ = camera;
    label_ = label;
    figures_ = figures;
    colors_ = colors;
  }
  public boolean release(double x, double y, int rawx, int rawy) {
    if (next_ != null) {
      return next_.click(x,y,rawx,rawy);                           
    } else {
      return false;
    }
  }
  public boolean click(double x, double y, int rawx, int rawy) {
    boolean processed = false;

    // Figure out what the nearest joint to this click is
    if (verbose_) System.out.println("click at ("+x+","+y+")");

    double distance = 0.05;
    int whichFig = -1;
    int whichLine = -1;
    boolean whichEnd = false; // true = A end, false = B end
      
    for (int fig=0; fig<figures_.length; fig++) {
      Line2D[] figlines = camera_.transform(figures_[fig].getLines());
      for (int jj=0; jj<figlines.length; jj++) {
        figlines[jj].c = colors_[fig];
        //if (verbose_) System.out.println("Examine line ["+fig+"]["+jj+"] = ("+figlines[jj].a.x+","+figlines[jj].a.y+") -> ("+figlines[jj].b.x+","+figlines[jj].b.y+")");
        double d = figlines[jj].a.distance(x, y);
        if (d < distance) {
          whichFig = fig;
          whichLine = jj;
          whichEnd = true;
          distance = d;
        }
        d = figlines[jj].b.distance(x, y);
        if (d < distance) {
          whichFig = fig;
          whichLine = jj;
          whichEnd = false;
          distance = d;
        }
      }
    }
    if (whichFig != -1) {
      int jnum = FigurePosition.lineToJoint(whichLine, whichEnd);
      if (verbose_) System.out.println("Closest to figure "+whichFig+", line "+whichLine+", end ="+(whichEnd?"A":"B")+", joint="+jnum);
      int line = FigurePosition.jointToLine(jnum);
      int dof = FigurePosition.jointToDof(jnum);
      Line3D[] figlines3d = figures_[whichFig].getLines();
      if (verbose_) System.out.println("Map back to end A of line "+line+" of figure "+whichFig+" with "+dof+" degrees of freedom");

      label_.setText(FigurePosition.jointToName(jnum));
      // Remove any existing controls
      canvas_.setClickConsumer(this);
      canvas_.setControl(null);

      if (jnum != FigurePosition.POSITION) {
        if (verbose_) figlines3d[line].a.dump("Add highlight of point ");
        PointHighlightControl ptctrl = new PointHighlightControl(figlines3d[line].a, camera_);
        canvas_.setControl(ptctrl);
      }

      // Generate controls for the joint
      AngleController[] angleController = new AngleController[dof];
      int MARGIN=20;
      for (int ii=0; ii<dof; ii++) {
        SharedDouble d = figures_[whichFig].jointValue(jnum, ii);
        angleController[ii] = new AngleController(d, 
                                                  FigurePosition.jointMin(jnum,ii),
                                                  FigurePosition.jointMax(jnum,ii),
                                                  trigger_, 
                                                  MARGIN, 
                                                  MARGIN + ii*AngleController.SIZE); 
        if (verbose_) System.out.println("Insert angle controller for fig "+whichFig+" joint "+jnum+" dof "+ii);
        // Insert this into the chain of consumers
        angleController[ii].nextClick_ = canvas_.getClickConsumer();
        canvas_.setClickConsumer(angleController[ii]);
        angleController[ii].nextDrag_ = canvas_.getDragConsumer();
        canvas_.setDragConsumer(angleController[ii]);
        angleController[ii].nextControl_ = canvas_.getControl();
        canvas_.setControl(angleController[ii]);
      }
      // Special case for figure positioning
      if (jnum == FigurePosition.POSITION) {
        PositionController posController = new PositionController(figures_[whichFig].posx_,
                                                                  figures_[whichFig].posy_,
                                                                  figures_[whichFig].posz_,
                                                                  camera_, 
                                                                  trigger_,
                                                                  MARGIN,
                                                                  MARGIN + dof*AngleController.SIZE); 
        
        // Insert this into the chain of consumers
        posController.nextClick_ = canvas_.getClickConsumer();
        canvas_.setClickConsumer(posController);
        posController.nextControl_ = canvas_.getControl();
        canvas_.setControl(posController);
        // Add a control that highlights the point, but this one may move as the controls
        // are frobbed
        PointHighlightControl ptctrl = new PointHighlightControl(figures_[whichFig].posx_,
                                                                 figures_[whichFig].posy_,
                                                                 figures_[whichFig].posz_,
                                                                 camera_);
        ptctrl.nextControl_ = canvas_.getControl();
        canvas_.setControl(ptctrl);
      }
      trigger_.trigger();
      processed = true;
    }
    if (!processed && (next_ != null)) {
      processed = next_.click(x,y,rawx,rawy);                           
    }
    return processed;
  }
};