// Copyright (c) 2006 David Drysdale.  See license.txt for license info
import java.awt.*;

class PositionController implements ClickConsumer, Control {
  final static int SIZE = 50;
  final static int ARROWFRACTION = 6;
  final static int CLICKFRACTION = 4;
  final static double DELTA = 0.1;
  Camera camera_;
  Control nextControl_;
  ClickConsumer nextClick_;
  Triggerable trigger_;
  PointHighlightControl ptctrl_;
  SharedDouble valuex_;
  SharedDouble valuey_;
  SharedDouble valuez_;
  Line3D linex_;
  Line3D liney_;
  Line3D linez_;
  int left_;
  int top_;
  static boolean verbose_ = false;

  public PositionController(SharedDouble valx, SharedDouble valy, SharedDouble valz, 
                            Camera camera, Triggerable trigger, int x, int y) {
    valuex_ = valx;
    valuey_ = valy;
    valuez_ = valz;
    camera_ = camera;
    trigger_ = trigger;
    left_ = x;
    top_ = y;
    // Set up some axes for future reference
    linex_ = new Line3D();
    linex_.a.x = 0; linex_.a.y = 0; linex_.a.z = 0;
    linex_.b.x = 1; linex_.b.y = 0; linex_.b.z = 0;
    liney_ = new Line3D();
    liney_.a.x = 0; liney_.a.y = 0; liney_.a.z = 0;
    liney_.b.x = 0; liney_.b.y = 1; liney_.b.z = 0;
    linez_ = new Line3D();
    linez_.a.x = 0; linez_.a.y = 0; linez_.a.z = 0;
    linez_.b.x = 0; linez_.b.y = 0; linez_.b.z = 1;
  }
  private void drawAxis(Graphics g, Line2D dir, int x, int y, Color c) {
    g.setColor(c);
    int cx = x + (SIZE/2);
    int cy = y + (SIZE/2);
    double theta = Math.atan2(dir.b.y-dir.a.y, dir.b.x-dir.a.x);
    int dx = (int)(Math.cos(theta)*(double)(SIZE/2));
    int dy = (int)(Math.sin(theta)*(double)(SIZE/2));
    g.drawLine(cx,cy, cx+dx,cy+dy);
    g.drawLine(cx,cy, cx-dx,cy-dy);
    // Draw some arrowheads
    double a1 = theta + 3*Math.PI/4;
    double a2 = theta - 3*Math.PI/4;
    g.drawLine(cx+dx,cy+dy, 
               cx+dx+ (int)(Math.cos(a1)*(double)(SIZE/ARROWFRACTION)),
               cy+dy+ (int)(Math.sin(a1)*(double)(SIZE/ARROWFRACTION)));
    g.drawLine(cx+dx,cy+dy, 
               cx+dx+ (int)(Math.cos(a2)*(double)(SIZE/ARROWFRACTION)),
               cy+dy+ (int)(Math.sin(a2)*(double)(SIZE/ARROWFRACTION)));
    a1 = theta + 3*Math.PI/4 + Math.PI;
    a2 = theta - 3*Math.PI/4 + Math.PI;
    g.drawLine(cx-dx,cy-dy, 
               cx-dx+ (int)(Math.cos(a1)*(double)(SIZE/ARROWFRACTION)),
               cy-dy+ (int)(Math.sin(a1)*(double)(SIZE/ARROWFRACTION)));
    g.drawLine(cx-dx,cy-dy, 
               cx-dx+ (int)(Math.cos(a2)*(double)(SIZE/ARROWFRACTION)),
               cy-dy+ (int)(Math.sin(a2)*(double)(SIZE/ARROWFRACTION)));
  }
  public void display(Drawable2D c, Graphics g) {
    // Ask the camera for 2D directions for axes
    Line2D lx = camera_.transform(linex_);
    Line2D ly = camera_.transform(liney_);
    Line2D lz = camera_.transform(linez_);
    drawAxis(g, lx, left_, top_, Color.blue);
    drawAxis(g, ly, left_, top_ + SIZE, Color.white);
    drawAxis(g, lz, left_, top_ + 2*SIZE, Color.red);

    if (nextControl_ != null) {
      nextControl_.display(c,g);
    }
  }
  private boolean checkAxis(Line2D dir, int x, int y, int rawx, int rawy, SharedDouble val, String debugstr) {
    boolean processed = false;
    int cx = x + (SIZE/2);
    int cy = y + (SIZE/2);
    double theta = Math.atan2(dir.b.y-dir.a.y, dir.b.x-dir.a.x);
    int dx = (int)(Math.cos(theta)*(double)(SIZE/2));
    int dy = (int)(Math.sin(theta)*(double)(SIZE/2));
    // Check for +ve end
    if (verbose_) System.out.println("compare ("+rawx+","+rawy+") against +("+(cx+dx)+","+(cy+dy)+") -("+(cx-dx)+","+(cy-dy)+") for "+debugstr);
    int d2 = ((cx+dx)-rawx)*((cx+dx)-rawx) + ((cy+dy)-rawy)*((cy+dy)-rawy);
    if (d2 < ((SIZE*SIZE)/(CLICKFRACTION*CLICKFRACTION))) {
      if (verbose_) System.out.println("increment "+debugstr+" from "+val.doubleValue()+" to "+(val.doubleValue() + DELTA));
      val.setDoubleValue(val.doubleValue() + DELTA);
      processed = true;
    } else {
      // Check for -ve end
      d2 = ((cx-dx)-rawx)*((cx-dx)-rawx) + ((cy-dy)-rawy)*((cy-dy)-rawy);
      if (d2 < ((SIZE*SIZE)/(CLICKFRACTION*CLICKFRACTION))) {
      if (verbose_) System.out.println("decrement "+debugstr+" from "+val.doubleValue()+" to "+(val.doubleValue() - DELTA));
        val.setDoubleValue(val.doubleValue() - DELTA);
        processed = true;
      }
    }
    return processed;
  }
  public boolean click(double x, double y, int rawx, int rawy) {
    boolean processed = false;
    // Ask the camera for 2D directions for axes
    Line2D lx = camera_.transform(linex_);
    Line2D ly = camera_.transform(liney_);
    Line2D lz = camera_.transform(linez_);
    // See if the click is near any of the arrowheads
    processed |= checkAxis(lx, left_, top_, rawx, rawy, valuex_, "x");
    if (!processed) processed |= checkAxis(ly, left_, top_ + SIZE, rawx, rawy, valuey_, "y");
    if (!processed) processed |= checkAxis(lz, left_, top_ + 2*SIZE, rawx, rawy, valuez_, "z");

    if (processed) {
      // Trigger a regeneration of the frames and a redisplay
      trigger_.generateFrames();
      trigger_.trigger();
    }
    if (!processed && (nextClick_ != null)) {
      // Pass it on down the chain
      processed = nextClick_.click(x,y,rawx,rawy);     
    }
    return processed;
  }
  public boolean release(double x, double y, int rawx, int rawy) {
    if (nextClick_ != null) {
      return nextClick_.release(x,y,rawx,rawy);     
    } else {
      return false;
    }
  }
};